# imports - standard imports
import os.path as osp
import sqlite3

# imports - module imports
from pipupgrade.__attr__    import __name__ as NAME
from pipupgrade.util.string import strip
from pipupgrade.util.system import makedirs, read
from pipupgrade             import config, log

logger = log.get_logger()

IntegrityError      = sqlite3.IntegrityError
OperationalError    = sqlite3.OperationalError

def _get_queries(buffer):
    queries = [ ]
    lines   = buffer.split(";")

    for line in lines:
        line = strip(line)
        queries.append(line)

    return queries

class DB:
    def __init__(self, path, timeout = 10):
        self.path        = path
        self._connection = None
        self.timeout     = timeout

    @property
    def connected(self):
        _connected = bool(self._connection)
        return _connected

    def connect(self, bootstrap = True):
        if not self.connected:
            self._connection = sqlite3.connect(self.path,
                timeout = self.timeout
            )
            self._connection.row_factory = sqlite3.Row

    def query(self, *args, **kwargs):
        if not self.connected:
            self.connect()

        cursor  = self._connection.cursor()
        cursor.execute(*args, **kwargs)
        self._connection.commit()

        results = cursor.fetchall()
        results = [dict(result) for result in results]

        if len(results) == 1:
            results = results[0]

        cursor.close()

        return results

_CONNECTION = None

def get_connection(bootstrap = True, log = False):
    global _CONNECTION

    if not _CONNECTION:
        if log:
            logger.info("Establishing a DataBase connection...")

        basepath    = osp.join(osp.expanduser("~"), ".%s" % NAME)
        makedirs(basepath, exist_ok = True)

        abspath     = osp.join(basepath, "db.db")

        _CONNECTION = DB(abspath)
        _CONNECTION.connect()

        if bootstrap:
            if log:
                logger.info("Bootstrapping DataBase...")

            abspath = osp.join(config.PATH["DATA"], "bootstrap.sql")
            buffer  = read(abspath)

            queries = _get_queries(buffer)

            for query in queries:
                _CONNECTION.query(query)

    return _CONNECTION


# pylint: disable=E1101

# imports - compatibility imports
from pipupgrade             import _compat
from pipupgrade._compat     import zip, _is_python_version
from pipupgrade.util._dict  import dict_from_list

# imports - standard imports
import sys
import inspect
import collections
import itertools

def get_function_arguments(fn):
    # https://stackoverflow.com/a/2677263
    params  = dict()
    success = False

    if _compat.PY2:
        argspec_getter = inspect.getargspec
        success        = True
    if _compat.PYTHON_VERSION >= (3,0) and (3,4) <= _compat.PYTHON_VERSION:
        argspec_getter = inspect.getfullargspec
        success        = True

    if success:
        argspec   = argspec_getter(fn)
        params    = dict_from_list(argspec.args, argspec.defaults or [])

    if _compat.PYTHON_VERSION >= (3,5):
        signature  = inspect.signature(fn)
        parameters = signature.parameters

        params     = { k: v.default for k, v in _compat.iteritems(parameters) }

        success    = True

    if not success:
        raise ValueError("Unknown Python Version {} for fetching functional arguments.".format(sys.version))

    return params

def auto_typecast(value):
    str_to_bool = lambda x: { "True": True, "False": False, "None": None}[x]

    for type_ in (str_to_bool, int, float):
        try:
            return type_(value)
        except (KeyError, ValueError, TypeError):
            pass

    return value


# imports - standard imports
import os, os.path as osp
import subprocess  as sp
from   distutils.spawn import find_executable

# imports - test imports
import pytest

# imports - module imports
from pipupgrade.util.system import (read, write, popen, which, makedirs,
    environment, touch)

def test_read(tmpdir):
    directory = tmpdir.mkdir("tmp")
    tempfile  = directory.join("foobar.txt")
    tempfile.write("foobar")

    assert tempfile.read() == read(str(tempfile))

    tempfile  = directory.join("barfoo.txt")
    tempfile.write(\
        """
        foobar
        \n
        barfoo
        """
    )

    assert tempfile.read() == read(str(tempfile))

def test_write(tmpdir):
    directory   = tmpdir.mkdir("tmp")
    tempfile    = directory.join("foobar.txt")
    
    path        = str(tempfile) 
    
    prev, next_ = "foobar", "barfoo"

    write(path, prev)
    assert tempfile.read() == prev

    write(path, next_)
    assert tempfile.read() == prev

    write(path, next_, force = True)
    assert tempfile.read() == next_

def test_popen(tmpdir):
    directory = tmpdir.mkdir("tmp")
    dirpath   = str(directory)

    string    = "Hello, World!"

    code, out, err = popen("echo %s" % string,
        output = True)
    assert code == 0
    assert out  == string
    assert not err
    
    env = dict({ "FOOBAR": "foobar" })
    code, out, err = popen("echo $FOOBAR; echo $PATH",
        output = True, env = env)
    assert code == 0
    assert out  == "%s\n%s" % (env["FOOBAR"], os.environ["PATH"])
    assert not err

    with pytest.raises(sp.CalledProcessError):
        code = popen("exit 42")

    errstr = "foobar"
    code, out, err = popen('python -c "raise Exception("%s")"' % errstr,
        output = True, raise_err = False)
    assert code == 1
    assert not out
    assert errstr in err

    filename = "foobar.txt"
    popen("touch %s" % filename, cwd = dirpath)
    assert osp.exists(osp.join(dirpath, filename))

    code = popen("echo $FOOBAR; echo $PATH", quiet = True)
    assert code == 0

def test_which():
    assert which("foobar") == None
    assert which("python") == find_executable("python")

    with pytest.raises(ValueError) as e:
        which("foobar", raise_err = True)

def test_makedirs(tmpdir):
    directory = tmpdir.mkdir("tmp")
    path      = osp.join(str(directory), "foo", "bar")

    makedirs(path)
    assert osp.exists(path)

    makedirs(path, exist_ok = True)
    assert osp.exists(path)

    with pytest.raises(OSError):
        makedirs(path)

def test_environment():
    details = environment()

    assert all((k in details for k in ("python_version", "os")))

    return details

def test_touch(tmpdir):
    directory = tmpdir.mkdir("tmp")
    path      = osp.join(str(directory), "foo")

    assert not osp.exists(path)

    touch(path)
    assert osp.exists(path)


# imports - standard imports
import sys, os, os.path as osp
import re
import json
import multiprocessing as mp
from   functools import partial

# imports - module imports
from pipupgrade.commands.helper import (
	get_registry_from_requirements,
	get_registry_from_pip,
	update_pip,
	update_pipfile,
	update_registry,
	_DEPENDENCY_FORMATS
)
from pipupgrade.model           import Project
from pipupgrade.model.project 	import get_included_requirements
from pipupgrade.commands.util 	import cli_format
from pipupgrade.util.array    	import flatten, sequencify
from pipupgrade.util.system   	import read, write, popen, which, environment
from pipupgrade.util.environ  	import getenvvar
from pipupgrade.util.datetime 	import get_timestamp_str
from pipupgrade 		      	import (_pip, request as req, cli,
	log, parallel
)
from pipupgrade._compat			import builtins
from pipupgrade.__attr__      	import __name__

logger = log.get_logger(level = log.DEBUG)

@cli.command
def command(
	pip_path          		 	= [ ],
	requirements 				= [ ],
	pipfile            			= [ ],
	project      				= None,
	pull_request 				= False,
	git_username 				= None,
	git_email    				= None,
	github_access_token			= None,
	github_reponame   		 	= None,
	github_username   			= None,
	target_branch    			= "master",
	latest						= False,
	format						= "table",
	all							= False,
	pip							= False,
	self 		 				= False,
	jobs						= 1,
	user		 				= False,
	check		 				= False,
	interactive  				= False,
	yes			 				= False,
	no_included_requirements 	= False,
	no_cache		            = False,
	no_color 	 				= True,
	output						= None,
	force						= False,
	verbose		 				= False
):
	if not verbose:
		logger.setLevel(log.NOTSET)
		
	logger.info("Environment: %s" % environment())

	cli.echo(cli_format("Checking...", cli.YELLOW))
	logger.info("Arguments Passed: %s" % locals())

	pip_path    = pip_path or [ ]
	pip_path    = [which(p) for p in pip_path] or _pip._PIP_EXECUTABLES

	logger.info("`pip` executables found: %s" % pip_path)
	
	logger.info("Using %s jobs..." % jobs)

	registries  = [ ]

	if pip:
		logger.info("Updating pip executables: %s" % " ".join(pip_path))

		with parallel.no_daemon_pool(processes = jobs) as pool:
			pool.map(
				partial(
					update_pip, **{ "user": user, "quiet": not verbose }
				),
				pip_path
			)

	if self:
		package = __name__
		logger.info("Updating %s..." % package)

		cli.echo(cli_format("Updating %s..." % package, cli.YELLOW))

		_pip.call("install", package, user = user, quiet = not verbose,
			no_cache = True, upgrade = True)

		cli.echo("%s upto date." % cli_format(package, cli.CYAN))
	else:
		if project:
			project		 = sequencify(project)
			requirements = requirements or [ ]
			pipfile      = pipfile      or [ ]

			logger.info("Detecting projects and its dependencies...")
			
			with parallel.no_daemon_pool(processes = jobs) as pool:
				project       = pool.map(
					partial(
						Project.from_path,
						**{ "recursive_search": force }
					),
					project
				)

				requirements += flatten(map(lambda p: p.requirements, project))
				pipfile      += flatten(map(lambda p: [p.pipfile] if p.pipfile else [], project))
			
			logger.info("Updating projects %s..." % project)

		if requirements:
			logger.info("Detecting requirements...")

			if not no_included_requirements:
				with parallel.no_daemon_pool(processes = jobs) as pool:
					results       = pool.map(get_included_requirements,
						requirements)
					requirements += flatten(results)

			logger.info("Requirements found: %s" % requirements)
			
			with parallel.no_daemon_pool(processes = jobs) as pool:
				results       = pool.map(
					partial(
						get_registry_from_requirements,
						**{ "sync": no_cache, "jobs": jobs }
					),
					requirements
				)
				registries    += results
		else:
			with parallel.no_daemon_pool(processes = jobs) as pool:
				results       = pool.map(
					partial(
						get_registry_from_pip,
						**{ "user": user, "sync": no_cache,
							"outdated": not all,
							"build_dependency_tree": format in _DEPENDENCY_FORMATS,
							"jobs": jobs
						}
					),
					pip_path
				)

				registries    += results

		logger.info("Updating registries: %s..." % registries)

		# TODO: Tweaks within parallel.no_daemon_pool to run serially.
		if yes:
			with parallel.no_daemon_pool(processes = jobs) as pool:
				results = pool.map(
					partial(
						update_registry,
						**{ "yes": yes, "user": user, "check": check,
							"latest": latest, "interactive": interactive,
							"verbose": verbose, "format_": format, "all": all }
					),
					registries
				)
		else:
			for registry in registries:
				update_registry(registry, yes = yes, user = user, check = check,
					latest = latest, interactive = interactive, verbose = verbose,
					format_ = format, all = all
				)

		if pipfile:
			logger.info("Updating Pipfiles: %s..." % pipfile)

			cli.echo(cli_format("Updating Pipfiles: %s..." % ", ".join(pipfile), cli.YELLOW))

			with parallel.no_daemon_pool(processes = jobs) as pool:
				results = pool.map(
					partial(
						update_pipfile,
						**{ "verbose": verbose }
					),
					pipfile
				)

				if builtins.all(results):
					cli.echo(cli_format("Pipfiles upto date.", cli.GREEN))

		if project and pull_request:
			errstr = '%s not found. Use %s or the environment variable "%s" to set value.'

			if not git_username:
				raise ValueError(errstr % ("Git Username", "--git-username", getenvvar("GIT_USERNAME")))
			if not git_email:
				raise ValueError(errstr % ("Git Email",    "--git-email",    getenvvar("GIT_EMAIL")))
			
			for p in project:
				popen("git config user.name  %s" % git_username, cwd = p.path)
				popen("git config user.email %s" % git_email,    cwd = p.path)

				_, output, _ = popen("git status -s", output = True)

				if output:
					branch   = get_timestamp_str(format_ = "%Y%m%d%H%M%S")
					popen("git checkout -B %s" % branch, quiet = not verbose)

					title    = "fix(dependencies): Update dependencies to latest"
					body     = ""

					# TODO: cross-check with "git add" ?
					files    = p.requirements + [p.pipfile]
					popen("git add %s" % " ".join(files), quiet = not verbose, cwd = p.path)
					popen("git commit -m '%s'" % title, quiet = not verbose, cwd = p.path)

					popen("git push origin %s" % branch, quiet = not verbose, cwd = p.path)

					if not github_reponame:
						raise ValueError(errstr % ("GitHub Reponame", "--github-reponame", getenvvar("GITHUB_REPONAME")))
					if not github_username:
						raise ValueError(errstr % ("GitHub Username", "--github-username", getenvvar("GITHUB_USERNAME")))

					url       = "/".join(["https://api.github.com", "repos", github_username, github_reponame, "pulls"])
					headers   = dict({
						 "Content-Type": "application/json",
						"Authorization": "token %s" % github_access_token
					})
					data      = dict(
						head  = "%s:%s" % (git_username, branch),
						base  = target_branch,
						title = title,
						body  = body
					)
					
					# Although there's monkey patch support for the "requests"
					# library, avoid using the "json" parameter which was
					# added in requests 2.4.2+
					response  = req.post(url, data = json.dumps(data), headers = headers)

					if response.ok:
						response = response.json()
						number   = response["number"]

						url      = "/".join(map(str, ["https://github.com", github_username, github_reponame, "pull", number]))
						message  = "Created a Pull Request at %s" % url

						cli.echo(cli_format(message, cli.GREEN))
					else:
						response.raise_for_status()



import os.path as osp

# imports - compatibility imports
from pipupgrade.commands    import command
from pipupgrade.util._dict  import merge_dict

# imports - test imports
import pytest

# imports - test imports
from testutils import mock_input, PATH

def test_command_self(capfd):
    command(self = True)
    out, err = capfd.readouterr()
    assert "upto date." in out


# imports - module imports
from pipupgrade.util._dict import (
    merge_dict,
    dict_from_list
)

def test_merge_dict():
    assert merge_dict({ "foo": "bar" }, { "bar": "baz" })     == { "foo": "bar", "bar": "baz" }
    assert merge_dict({ "foo": "bar" }, { "foo": "baz" })     == { "foo": "baz" }
    assert merge_dict({ 1: 2 }, { 3: 4 }, { 5: 6 }, { 7: 8 }) == { 1: 2, 3: 4, 5: 6, 7: 8 }
    assert merge_dict({ 1: 2 }, { 1: 3 }, { 1: 4 }, { 1: 1 }) == { 1: 1 }

def test_dict_from_list():
    assert dict_from_list(["foo", "bar"], [1, 2]) == dict(foo = 1, bar = 2)
    assert dict_from_list([1, 2], ["foo", "bar"]) == { 1: "foo", 2: "bar" }


# imports - module imports
from pipupgrade import parallel

def test_pool():
    def _assert(pool):
        results = pool.map(sum, [(1, 2), (3, 4)])

        assert results[0] == 3
        assert results[1] == 7

    with parallel.pool() as pool:
        _assert(pool)

    with parallel.pool(class_ = parallel.NoDaemonProcessPool) as pool:
        _assert(pool)

    with parallel.no_daemon_pool() as pool:
        _assert(pool)


# imports - module imports
from pipupgrade._compat    import StringIO, input

# imports - test imports
from testutils import mock_input, assert_input

def test_mock_input(capfd):
    query = "foobar"

    assert_input(capfd, query, "Y")
    assert_input(capfd, query, "Y", input_ = input)


# imports - standard imports
import os, os.path as osp
import fnmatch
import glob

# imports - module imports
from pipupgrade import log
from pipupgrade.util.string import strip

logger = log.get_logger()

class Project:
	@staticmethod
	def from_path(path, *args, **kwargs):
		project = Project(path, *args, **kwargs)
		return project

	def __init__(self, path, recursive_search = False):
		path = osp.realpath(path)

		if not osp.exists(path):
			logger.error("Failed to initialize project for path %s." % path)
			raise ValueError("Path %s does not exist." % path)
		
		logger.info("Initializing project with path %s." % path)
		self.path         = path
		logger.info("Fetching requirements...")
		self.requirements = self._get_requirements(
			recursive_search = recursive_search)

		logger.info("Fetching Pipfile...")
		self.pipfile      = self._get_pipfile()

	def _get_requirements(self, recursive_search = False):
		# COLLECT ALL THE REQUIREMENTS FILES!
		path         = self.path
		requirements = [ ]

		# Detect Requirements Files
		level = 0
		for root, _, files in os.walk(path):
			for file_ in fnmatch.filter(files, "requirements*.txt"):
				requirement = osp.join(root, file_)
				logger.info("Requirement found at %s." % requirement)
				requirements.append(requirement)

			if level == 0 and not recursive_search:
				break
			else:
				level += 1
	
		# Check if requirements is a directory
		if osp.isdir(osp.join(path, "requirements")):
			for requirement in glob.glob(osp.join(path, "requirements", "*.txt")):
				logger.info("Requirement found at %s." % requirement)
				requirements.append(requirement)

		return requirements

	def _get_pipfile(self):
		path    = self.path
		pipfile = osp.join(path, "Pipfile")

		if pipfile:
			logger.info("Pipfile found at %s." % pipfile)
		
		return pipfile if osp.exists(pipfile) else None

	def __repr__(self):
		repr_ = "<Project %s>" % self.path
		return repr_

def get_included_requirements(filename):
	path         = osp.realpath(filename)
	basepath     = osp.dirname(path)
	requirements = [ ]

	with open(path) as f:
		content = f.readlines()

		for line in content:
			line = strip(line)

			if line.startswith("-r "):
				filename = line.split("-r ")[1]
				realpath = osp.join(basepath, filename)
				requirements.append(realpath)

				requirements += get_included_requirements(realpath)

	return requirements


# imports - standard imports
import sys, os, os.path as osp
import logging

def pardir(fname, level = 1):
    fname = osp.realpath(fname)

    for _ in range(level):
        fname = osp.dirname(fname)
    return fname

PATH          = dict()
PATH["BASE"]  = pardir(__file__, 2)
PATH["TESTS"] = osp.join(PATH["BASE"], "tests")

sys.path.append(
    osp.join(PATH["TESTS"], "helpers")
)


# imports - standard imports
import logging

# imports - module imports
from pipupgrade.__attr__   import __name__ as NAME

NOTSET      = logging.NOTSET
DEBUG       = logging.DEBUG
INFO        = logging.INFO
WARNING     = logging.WARNING
ERROR       = logging.ERROR
CRITICAL    = logging.CRITICAL

_FORMAT     = '%(asctime)s | %(levelname)s | %(message)s'
_LOGGER     = None

def get_logger(name = NAME, level = DEBUG, format_ = _FORMAT):
    global _LOGGER

    if not _LOGGER:
        formatter = logging.Formatter(format_)

        handler   = logging.StreamHandler()
        handler.setFormatter(formatter)

        logger    = logging.getLogger(name)
        logger.setLevel(level)

        logger.addHandler(handler)
        
        _LOGGER   = logger
    
    return _LOGGER


# imports - standard imports
import re
from   functools import partial

# imports - module imports
from pipupgrade.model.package   import Package, _get_pip_info
from pipupgrade                 import _pip, parallel
from pipupgrade.util.array      import compact, flatten
from pipupgrade.util.string     import kebab_case
from pipupgrade._compat		    import iteritems, iterkeys, itervalues
from pipupgrade.tree            import Node as TreeNode
from pipupgrade.log             import get_logger

logger = get_logger()

# cache package information
_INFO_DICT = dict()
# cache dependency trees
_TREE_DICT = dict()

def _build_packages_info_dict(packages, pip_exec = None):
    details         = _get_pip_info(*packages, pip_exec = pip_exec)

    requirements    = [ ]

    for name, detail in iteritems(details):
        if not name in _INFO_DICT:
            _INFO_DICT[name] = dict({
                     "version": detail["version"], 
                "dependencies": compact(detail["requires"].split(", "))
            })

            for requirement in _INFO_DICT[name]["dependencies"]:
                if requirement not in requirements:
                    requirements.append(requirement)

    if requirements:
        _build_packages_info_dict(requirements, pip_exec = pip_exec)

def _create_package(name, sync = False):
    package                 = Package(name, sync = sync)
    package.current_version = _INFO_DICT[name]["version"]
    
    return package

def _get_dependency_tree_for_package(package, parent = None, sync = False,
    jobs = 1):
    if package.name not in _TREE_DICT:
        logger.info("Building dependency tree for package: %s..." % package)

        tree            = TreeNode(package, parent = parent)

        dependencies    = [ ]
        
        with parallel.no_daemon_pool(processes = jobs) as pool:
            dependencies = pool.map(
                partial(
                    _create_package, **{
                        "sync": sync
                    }
                ),
                _INFO_DICT[package.name]["dependencies"]
            )

        with parallel.no_daemon_pool(processes = jobs) as pool:
            children = pool.map(
                partial(
                    _get_dependency_tree_for_package, **{
                        "parent": tree
                    }
                ),
                dependencies
            )

            if children:
                tree.add_children(*children)

        _TREE_DICT[package.name] = tree
    else:
        logger.info("Using cached dependency tree for package: %s." % package)

    tree        = _TREE_DICT[package.name]
    tree.parent = parent

    return tree

class Registry:
    def __init__(self,
        source,
        packages                = [ ],
        installed               = False,
        sync                    = False,
        build_dependency_tree   = False,
        jobs                    = 1
    ):
        self.source = source

        args        = { "sync": sync }

        if installed:
            args["pip_exec"] = source
        
        with parallel.no_daemon_pool(processes = jobs) as pool:
            self.packages = pool.map(partial(Package, **args), packages)

        self.installed = installed
        
        if installed and build_dependency_tree and self.packages:
            self._build_dependency_tree_for_packages(sync = sync, jobs = jobs)

    def _build_dependency_tree_for_packages(self, sync = False, jobs = 1):
        names = [p.name for p in self.packages]
        _build_packages_info_dict(names, pip_exec = self.source)

        for package in self.packages:
            package.dependency_tree = _get_dependency_tree_for_package(package,
                sync = sync, jobs = 1)


# imports - module imports
from pipupgrade       import cli
from pipupgrade.table import _sanitize_string, Table

def test__sanitize_string():
    assert _sanitize_string(cli.format("foobar", cli.GREEN)) == "foobar"
    assert _sanitize_string(cli.format("foobar", cli.BOLD))  == "foobar"

def test_table():
    table  = Table()
    assert table.empty
    
    dummy  = ["foo", "bar"]

    table.insert(dummy)
    assert not table.empty
    
    string = table.render()
    assert string.count("\n") == 1

    table.header = dummy
    string = table.render()
    assert string.count("\n") == 2

    table.insert(dummy)
    string = table.render()
    assert string.count("\n") == 3


# imports - compatibility imports
from pipupgrade.util.system import popen, pardir
from pipupgrade.util.string import safe_decode
from pipupgrade._compat     import StringIO, EX_OK, input

# imports - standard imports
import sys
import os.path as osp
from   contextlib import contextmanager
import traceback

PATH            = dict()
PATH["BASE"]    = pardir(__file__, 2)
PATH["DATA"]    = osp.join(PATH["BASE"], "data") 

__STDIN__ = sys.stdin

@contextmanager
def mock_input(args):
    # https://stackoverflow.com/a/36491341
    sys.stdin = args
    yield
    sys.stdin = __STDIN__

def assert_stdout(capfd, output):
    stdout, _ = capfd.readouterr()
    assert safe_decode(output) == safe_decode(stdout)

def assert_input(capfd, text, output, expected = None, input_ = None, stdout = None, input_args = { }):
    if expected == None:
        expected = output
    input_ = input_ or input
    stdout = stdout or text
    
    with mock_input(StringIO(output)):
        assert input_(text, *input_args) == expected
        assert_stdout(capfd, stdout)

class CLIRunnerResult:
    pass

class Capture:
    def __enter__(self):
        self._stdout    = sys.stdout
        
        self._output    = StringIO()

        sys.stdout      = self._output

        return self
        
    def __exit__(self, *args):
        self.output     = safe_decode(self._output.getvalue())
        del self._output
        sys.stdout      = self._stdout

class CLIRunner:
    def invoke(self, command, args, **kwargs):
        code    = EX_OK
        error   = ""

        with Capture() as capture:
            try:
                command(**args)
            except Exception:
                code    = 1
                error   = traceback.format_exception()
        
        result          = CLIRunnerResult()

        result.code     = code
        result.output   = capture.output
        result.eror     = error

        return result


# imports - test imports
import pytest

# imports - module imports
from   pipupgrade.util.datetime import (get_timestamp_str, check_datetime_format,
    _DEFAULT_TIMESTAMP_FORMAT)

def test_get_timestamp_str():
    assert check_datetime_format(get_timestamp_str(), _DEFAULT_TIMESTAMP_FORMAT)

    format_ = '%Y-%m-%d'
    assert     check_datetime_format(get_timestamp_str(format_), format_)
    assert not check_datetime_format(get_timestamp_str(),        format_)

    with pytest.raises(ValueError):
        assert check_datetime_format(get_timestamp_str(), format_, raise_err = True)


# imports - module imports
from pipupgrade.util.array import (
    compact,
    squash,
    flatten,
    sequencify
)

# imports - test imports
import pytest

def test_compat():
    assert compact([1, 2, None])                        == [1, 2]
    assert compact([1, 2])                              == [1, 2]
    assert compact([1, 2, "", "foo"])                   == [1, 2, "foo"]
    assert compact(["foo", "bar", ""], type_ = tuple)   == ("foo", "bar")

def test_squash():
    assert squash(["foo"])          == "foo"
    assert squash(["foo", "bar"])   == ["foo", "bar"]

def test_flatten():
    assert flatten([[1, 2], [3, 4]])    == [1, 2, 3, 4]
    assert flatten([[1, 2]])            == [1, 2]
    assert flatten([[1, 2], [ ]])       == [1, 2]

    with pytest.raises(TypeError):
        assert flatten([[1, 2], None])

def test_sequencify():
    assert sequencify("foobar") == ["foobar"]
    assert sequencify([1,2,3])  == [1,2,3]
    assert sequencify([1,2,3])  != [3,2,1]
    assert sequencify([])       == []
    assert sequencify(None)     == [None]

    assert sequencify("foobar", type_ = tuple) == ("foobar",)
    assert sequencify([1,2,3],  type_ = tuple) == (1,2,3)
    assert sequencify([1,2,3],  type_ = tuple) != (3,2,1)
    assert sequencify([],       type_ = tuple) == tuple()
    assert sequencify(None,     type_ = tuple) == (None,)


# imports - test imports
import pytest

# imports - module imports
from pipupgrade.util.environ import getenvvar, getenv, value_to_envval

def test_getenvvar():
    assert getenvvar("FOOBAR")                  == "PIPUPGRADE_FOOBAR"
    assert getenvvar("FOOBAR", prefix = False)  == "FOOBAR"

def test_getenv():
    with pytest.raises(KeyError):
        assert getenv("ABCDEFGHIJKLMNOPQRSTUVWZYX", raise_err = True)

def test_value_to_envval():
    assert value_to_envval(True)     == "true"
    assert value_to_envval(False)    == "false"
    assert value_to_envval(12345)    == "12345"
    assert value_to_envval("foobar") == "foobar"

    with pytest.raises(TypeError):
        value_to_envval(str)
        
    with pytest.raises(TypeError):
        value_to_envval([ ])


# imports - standard imports
import os
import argparse
import multiprocessing as mp

# imports - module imports
from pipupgrade.__attr__     import (
    __name__,
    __version__,
    __description__,
    __command__
)
from pipupgrade.util.environ  import getenv
from pipupgrade.cli           import util as _cli
from pipupgrade.cli.formatter import ArgumentParserFormatter
from pipupgrade.cli.util      import _CAN_ANSI_FORMAT
from pipupgrade._pip          import _PIP_EXECUTABLES

_DESCRIPTION_JUMBOTRON = \
"""
%s (v %s)

%s
""" % (
    _cli.format(__name__,        _cli.RED),
    _cli.format(__version__,     _cli.BOLD),
    _cli.format(__description__, _cli.BOLD)
)

def get_parser():
    parser = argparse.ArgumentParser(
        prog            = __command__,
        description     = _DESCRIPTION_JUMBOTRON,
        add_help        = False,
        formatter_class = ArgumentParserFormatter
    )
    parser.add_argument("--pip-path",
        action  = "append",
        help    = "Path to pip executable to be used."
    )
    parser.add_argument("-y", "--yes",
        action  = "store_true",
        default = getenv("ACCEPT_ALL_DIALOGS", False),
        help    = "Confirm for all dialogs."
    )
    parser.add_argument("-c", "--check",
        action  = "store_true",
        default = getenv("DRY_RUN", False),
        help    = "Perform a dry-run, avoid updating packages."
    )
    parser.add_argument("-l", "--latest",
        action  = "store_true",
        default = getenv("UPDATE_LATEST", False),
        help    = "Update all packages to latest."
    )
    parser.add_argument("-f", "--format",
        choices = ["table", "tree", "json", "yaml"],
        help    = "Display packages format.",
        default = getenv("DISPLAY_FORMAT", "table")
    )
    parser.add_argument("-a", "--all",
        action  = "store_true",
        default = getenv("DISPLAY_ALL_PACKAGES", False),
        help    = "List all packages."
    )
    parser.add_argument("--pip",
        action  = "store_true",
        default = getenv("UPDATE_PIP", False),
        help    = "Update pip."
    )
    parser.add_argument("-s", "--self",
        action  = "store_true",
        help    = "Update %s." % __name__
    )
    parser.add_argument("-r", "--requirements",
        action  = "append",
        help    = "Path(s) to requirements.txt file."
    )
    parser.add_argument("--pipfile",
        action  = "append",
        help    = "Path(s) to Pipfile"
    )
    parser.add_argument("-i", "--interactive",
        action  = "store_true",
        default = getenv("INTERACTIVE", False),
        help    = "Interactive Mode."
    )
    parser.add_argument("-p", "--project",
        action  = "append",
        help    = "Path(s) to Project"
    )
    parser.add_argument("--git-username",
        help    = "Git Username",
        default = getenv("GIT_USERNAME")
    )
    parser.add_argument("--git-email",
        help    = "Git Email",
        default = getenv("GIT_EMAIL")
    )
    parser.add_argument("--pull-request",
        action  = "store_true",
        help    = "Perform a Pull Request."
    )
    parser.add_argument("--github-access-token",
        help    = "GitHub Access Token",
        default = getenv("GITHUB_ACCESS_TOKEN")
    )
    parser.add_argument("--github-reponame",
        help    = "Target GitHub Repository Name",
        default = getenv("GITHUB_REPONAME")
    )
    parser.add_argument("--github-username",
        help    = "Target GitHub Username",
        default = getenv("GITHUB_USERNAME")
    )
    parser.add_argument("--target-branch",
        help    = "Target Branch",
        default = getenv("TARGET_BRANCH", "master")
    )
    parser.add_argument("-j", "--jobs",
        type    = int,
        help    = "Number of Jobs to be used.",
        default = getenv("JOBS", mp.cpu_count())
    )
    parser.add_argument("-u", "--user",
        action  = "store_true",
        default = getenv("USER_ONLY", False),
        help    = "Install to the Python user install directory for environment \
                    variables and user configuration."
    )
    parser.add_argument("--no-included-requirements",
        action  = "store_true",
        default = getenv("NO_INCLUDED_REQUIREMENTS", False),
        help    = "Avoid updating included requirements."
    )
    parser.add_argument("--no-cache",
        action  = "store_true",
        default = getenv("NO_CACHE", False),
        help    = "Avoid fetching latest updates from PyPI server."
    )
    parser.add_argument("-o", "--output",
        default = getenv("OUTPUT_FILE"),
        help    = "Print Output to File."
    )
    parser.add_argument("--force",
        action  = "store_true",
        default = getenv("FORCE", False),
        help    = "Force search for files within a project."
    )

    if _CAN_ANSI_FORMAT:
        parser.add_argument("--no-color",
            action  = "store_true",
            default = getenv("NO_COLOR", False),
            help    = "Avoid colored output."
        )

    parser.add_argument("-V", "--verbose",
        action  = "store_true",
        help    = "Display verbose output."
    )
    parser.add_argument("-v", "--version",
        action  = "version",
        version = __version__,
        help    = "Show %s's version number and exit." % __name__
    )
    parser.add_argument("-h", "--help",
        action  = "help",
        default = argparse.SUPPRESS,
        help    = "Show this help message and exit."
    )

    return parser

def get_args(args = None, known = True, as_dict = True):
    parser  = get_parser()

    if known:
        args, _ = parser.parse_known_args(args)
    else:
        args    = parser.parse_args(args)

    if as_dict:
        args = args.__dict__
        
    return args


# imports - module imports
from pipupgrade.cli.util   import *
from pipupgrade.cli.parser import get_args
from pipupgrade.util._dict import merge_dict
from pipupgrade.util.types import get_function_arguments

def command(fn):
    args    = get_args()
    
    params  = get_function_arguments(fn)

    params  = merge_dict(params, args)
    
    def wrapper(*args, **kwargs):
        return fn(**params)

    return wrapper


# imports - compatibility imports
from __future__ import absolute_import

# imports - standard imports
import time
import datetime as dt

_DEFAULT_TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S'

def get_timestamp_str(format_ = _DEFAULT_TIMESTAMP_FORMAT):
    now       = time.time()
    
    datetime_ = dt.datetime.fromtimestamp(now)
    string    = datetime_.strftime(format_)

    return string

def check_datetime_format(datetime, format_, raise_err = False):
    try:
        dt.datetime.strptime(datetime, format_)
    except ValueError:
        if raise_err:
            raise ValueError("Incorrect datetime format, expected %s" % format_)
        else:
            return False
    
    return True


# imports - standard imports
import sys, os, os.path as osp
import re
import json

# imports - module imports
from pipupgrade.model         	import Registry
from pipupgrade.commands.util 	import cli_format
from pipupgrade.table      	  	import Table
from pipupgrade.tree			import Node as TreeNode
from pipupgrade.util.string     import pluralize, strip
from pipupgrade.util.system   	import read, write, popen, which
from pipupgrade.util.array		import squash
from pipupgrade 		      	import (_pip, cli, semver,
	log, parallel
)

logger = log.get_logger()

_SEMVER_COLOR_MAP 	= dict(
	major = cli.RED,
	minor = cli.YELLOW,
	patch = cli.GREEN
)
_DEPENDENCY_FORMATS = ("tree", "json", "yaml")

def _cli_format_semver(version, type_):
	def _format(x):
		return cli_format(x, cli.YELLOW)
	
	try:
		semver.parse(version)
		
		if type_ == "major":
			version    = _format(version)
		if type_ == "minor":
			index      = version.find(".", 1) + 1
			head, tail = version[:index], version[index:]
			version    = "".join([head, _format(tail)])
		if type_ == "patch":
			index      = version.find(".", 2) + 1
			head, tail = version[:index], version[index:]
			version    = "".join([head, _format(tail)])
	except ValueError:
		pass

	return version

def _update_requirements(path, package):
	path 	= osp.realpath(path)
	
	content = read(path)
		
	try:
		pattern = r"{package}(=={version})*".format(
			package = re.escape(package.name),
			version = re.escape(package.current_version)
		)
		lines   = content.splitlines()
		nlines  = len(lines)
		
		with open(path, "w") as f:
			for i, line in enumerate(lines):
				if re.search(pattern, line, flags = re.IGNORECASE):
					current_version = package.current_version

					if current_version:
						current_version = package.current_version.replace("==", "")

					line = line.replace(
						"==%s" % current_version,
						"==%s" % package.latest_version
					)
					
				f.write(line)

				if i < nlines - 1:
					f.write("\n")
	except Exception:
		# In case we fucked up!
		write(path, content, force = True)

def update_pip(pip_exec, user = None, quiet = None):
	cli.echo(cli_format("Updating %s..." % pip_exec, cli.YELLOW))

	output = _pip.call("install", "pip", pip_exec = pip_exec, 
		user = user, quiet = quiet, no_cache_dir = True, upgrade = True)

	if isinstance(output, int):
		code = output
	else:
		code = output[0]

	if not code:
		cli.echo("%s upto date." % cli_format(pip_exec, cli.CYAN))

	return output

def update_pipfile(pipfile, verbose = False):
	if not verbose:
		logger.setLevel(log.NOTSET)

	realpath = osp.realpath(pipfile)
	basepath = osp.dirname(realpath)

	logger.info("Searching for `pipenv`...")
	pipenv   = which("pipenv")

	if not pipenv:
		logger.info("Attempting to install pipenv...")

		_pip.call("install", "pipenv")

		pipenv = which("pipenv", raise_err = True)

	logger.info("`pipenv` found.")

	code     = popen("%s update" % pipenv, quiet = not verbose, cwd = basepath)

	return code == 0

def get_registry_from_requirements(requirements, sync = False, jobs = 1):
	path = osp.realpath(requirements)

	if not osp.exists(path):
		cli.echo(cli_format("{} not found.".format(path), cli.RED))
		sys.exit(os.EX_NOINPUT)
	else:
		packages =  _pip.parse_requirements(requirements, session = "hack")
		registry = Registry(source = path, packages = packages, sync = sync,
			jobs = jobs
		)

	logger.info("Packages within requirements %s found: %s..." % (
		requirements, registry.packages)
	)

	return registry

def get_registry_from_pip(pip_path, user = False, sync = False, outdated = True,
	build_dependency_tree = False, jobs = 1
):
	logger.info("Fetching installed packages for %s..." % pip_path)

	_, output, _ = _pip.call("list", user = user, outdated = outdated, \
		format = "json", pip_exec = pip_path, output = True)

	packages     = json.loads(output)
	logger.info("%s packages found for %s." % (len(packages), pip_path))
	registry     = Registry(source = pip_path, packages = packages,
		installed = True, sync = sync,
		build_dependency_tree = build_dependency_tree, jobs = jobs)

	logger.info("Packages within `pip` %s found: %s..." % (pip_path, registry.packages))
	# _pip.get_installed_distributions() # https://github.com/achillesrasquinha/pipupgrade/issues/13

	return registry

def _format_package(package):
	difference = None
	
	try:
		difference = semver.difference(package.current_version,
			package.latest_version)
	except (TypeError, ValueError):
		pass

	string = "* %s" % cli_format(package.name, _SEMVER_COLOR_MAP.get(difference, cli.CLEAR))

	if getattr(package, "has_dependency_conflict", False):
		string += cli_format(" [dependency conflict]", cli.RED)

	if package.current_version:
		string += " (%s)" % package.current_version

	if package.latest_version and package.current_version != package.latest_version:
		string += " -> (%s)" % _cli_format_semver(package.latest_version, difference)

	string += " " + cli_format("[%s]" % package.home_page, cli.CYAN)

	return string

def _render_dependency_tree(packages):
	rendered = [ ]

	for package in packages:
		dependencies 	= package.dependency_tree
		string			= dependencies.render(indent = 4,
			formatter = lambda package: _format_package(package)
		)

		sanitized		= strip(string)
		
		rendered.append(sanitized)

	string = strip("\n".join(rendered))

	return string

def _render_json(packages):
	dicts = [ ]

	for package in packages:
		dict_ = package.to_dict()
		dicts.append(dict_)
		
	return dicts

def _render_yaml(packages):
	try:
		import yaml

		content = _render_json(packages)
		dict_   = dict()

		for details in content:
			name = details.pop("name")
			dict_[name] = details

		string  = strip(yaml.dump(dict_))

		return string
	except ImportError:
		raise ValueError((
			"Unable to import yaml. "
			"Please install pyyaml. https://github.com/yaml/pyyaml."
		))

def _resolve_dependencies(nodes):
	for node in nodes:
		deptree = node.dependency_tree
		found 	= deptree.find(
			lambda x: x.parent != None and x.obj.difference == "major"
		)

		if found:
			node.has_dependency_conflict = True
	
	return nodes

def update_registry(registry,
	yes         = False,
	user 		= False,
	check	    = False,
	latest		= False,
	interactive = False,
	format_		= "table",
	all			= False,
	verbose 	= False):
	source   = registry.source
	packages = registry.packages
	
	table 	 = Table(header = ["Name", "Current Version", "Latest Version",
		"Home Page"])
	nodes	 = [ ]
	render   = False
	dinfo 	 = [ ] # Information DataFrame

	for package in packages:
		package.source    	= source
		package.installed 	= registry.installed
		
		current_version		= package.current_version
		
		if all or current_version:
			current_version = current_version.replace("==", "")

		if all or (package.latest_version and current_version != package.latest_version):
			render	  = True

			if format_ in _DEPENDENCY_FORMATS:
				nodes.append(package)
			else:
				table.insert([
					cli_format(package.name, _SEMVER_COLOR_MAP.get(package.difference, cli.CLEAR)),
					package.current_version or "na",
					_cli_format_semver(package.latest_version, package.difference),
					cli_format(package.home_page, cli.CYAN)
				])

			dinfo.append(package)

	stitle = "Installed Distributions (%s)" % source if registry.installed else source
	
	if render:
		if format_ in _DEPENDENCY_FORMATS:
			nodes  = _resolve_dependencies(nodes)
			dinfo  = nodes

		if 	 format_ == "tree":
			string = _render_dependency_tree(nodes)
		elif format_ == "json":
			string = _render_json(nodes)
		elif format_ == "yaml":
			string = _render_yaml(nodes)
		elif format_ == "table":
			string = table.render()
	
		cli.echo("\nSource: %s\n" % stitle)
		
		if not interactive or check:
			cli.echo(string)
			cli.echo()

		if not check:
			packages  = [p for p in dinfo if p.difference != "major" 
				or getattr(p, "has_dependency_conflict", False) or latest]
			npackages = len(packages)

			spackages = pluralize("package", npackages) # Packages "string"
			query     = "Do you wish to update %s %s?" % (npackages, spackages)

			if npackages and (yes or interactive or cli.confirm(query, quit_ = True)):
				for i, package in enumerate(packages):
					update = True
					
					query  = "%s (%s) -> (%s)" % (
						cli_format(package.name, _SEMVER_COLOR_MAP.get(package.difference, cli.CLEAR)),
						package.current_version,
						_cli_format_semver(package.latest_version, package.difference)
					)

					if interactive:
						update = yes or cli.confirm(query)
					
					if update:
						cli.echo(cli_format(
							"Updating %s of %s %s: %s" % (
								i + 1,
								npackages,
								spackages,
								cli_format(package.name, cli.GREEN)
							)
						, cli.BOLD))

						if not package.installed:
							_update_requirements(package.source, package)
						else:
							_pip.call("install", package.name,
								pip_exec = package.source, user = user,
								quiet = not verbose, no_cache_dir = True,
								upgrade = True
							)
	else:
		cli.echo("%s upto date." % cli_format(stitle, cli.CYAN))


# imports - compatibility imports
from pipupgrade._compat import iteritems, iterkeys

# imports - module imports
from pipupgrade import cli
from pipupgrade.cli import get_args
from pipupgrade.util._dict import merge_dict

def test_command():
    def _assert_command(values, override = dict(), initial = dict()):
        @cli.command
        def foobar(*args, **kwargs):
            args    = get_args()
            params  = merge_dict(args, override)
            
            for k, v in iteritems(values):
                assert params[k] == v

            if initial:
                for k in iterkeys(initial):
                    assert initial[k] == args[k]
        
        foobar()
    
    _assert_command(dict(yes    = False))
    _assert_command(dict(latest = True), dict(latest = True), dict(latest = False))


class HandlerRegistry(dict):
    def __missing__(self, name):
        if '.' not in name:
            handler = __import__(name)
        else:
            module_name, handler_name = name.rsplit('.', 1)

            module  = __import__(module_name, {}, {}, [handler_name])
            handler = getattr(module, handler_name)

        self[name] = handler

        return handler

_HANDLER_REGISTRY = HandlerRegistry()

def import_handler(name):
    handler = _HANDLER_REGISTRY[name]
    return handler


# imports - compatibility imports
from pipupgrade._compat import iteritems

# imports - standard imports
import os, os.path as osp
import errno
import platform
import subprocess  as sp
from   distutils.spawn import find_executable

# imports - module imports
from pipupgrade.exception   import PopenError
from pipupgrade.util.string import strip, safe_decode
from pipupgrade._compat     import iteritems
from pipupgrade.log         import get_logger

logger = get_logger()

def read(fname):
    with open(fname) as f:
        data = f.read()
    return data

def write(fname, data = None, force = False):
    if not osp.exists(fname) or force:
        with open(fname, "w") as f:
            if data:
                f.write(data)

def which(executable, raise_err = False):
    exec_ = find_executable(executable)
    
    if not exec_ and raise_err:
        raise ValueError("Executable %s not found." % exec_)
    
    return exec_

def pardir(fname, level = 1):
    for _ in range(level):
        fname = osp.dirname(fname)
    return fname

def popen(*args, **kwargs):
    output      = kwargs.get("output", False)
    quiet       = kwargs.get("quiet" , False)
    directory   = kwargs.get("cwd")
    environment = kwargs.get("env")
    shell       = kwargs.get("shell", True)
    raise_err   = kwargs.get("raise_err", True)

    environ     = os.environ.copy()
    if environment:
        environ.update(environment)

    for k, v in iteritems(environ):
        environ[k] = str(v)

    command     = " ".join([str(arg) for arg in args])

    logger.info("Executing command: %s" % command)

    if quiet:
        output  = True
    
    proc        = sp.Popen(command,
        bufsize = -1,
        stdin   = sp.PIPE if output else None,
        stdout  = sp.PIPE if output else None,
        stderr  = sp.PIPE if output else None,
        env     = environ,
        cwd     = directory,
        shell   = shell
    )

    code       = proc.wait()

    if code and raise_err:
        raise PopenError(code, command)

    if output:
        output, error = proc.communicate()

        if output:
            output = safe_decode(output)
            output = strip(output)

        if error:
            error  = safe_decode(error)
            error  = strip(error)

        if quiet:
            return code
        else:
            return code, output, error
    else:
        return code

def makedirs(dirs, exist_ok = False):
    try:
        os.makedirs(dirs)
    except OSError as e:
        if not exist_ok or e.errno != errno.EEXIST:
            raise

def environment():
    environ = dict()
    
    environ["python_version"]   = platform.python_version()
    environ["os"]               = platform.platform()

    return environ

def touch(filename):
    if not osp.exists(filename):
        with open(filename, "w") as f:
            pass


# imports - compatibility imports
from pipupgrade._compat import HTTPError

# imports - test imports
import pytest

# imports - standard imports
from pipupgrade import request as req

def test_get():
    res  = req.get("https://httpbin.org/get")
    assert res.ok
    assert res.status_code == 200
    
    json = res.json() 
    assert all(k in json for k in ("url", "origin", "headers", "args"))

    res.raise_for_status()
    
    res  = req.get("http://httpbin.org/status/404")
    assert not res.ok
    assert res.status_code == 404

    with pytest.raises(HTTPError):
        res.raise_for_status()

    assert str(res) == "<Response [{code}]>".format(
        code = 404
    )

def test_post():
    res  = req.post("https://httpbin.org/post")
    assert res.ok
    assert res.status_code == 200
    
    json = res.json() 
    assert all(k in json for k in ("url", "origin", "headers", "args"))

    res.raise_for_status()
    
    res  = req.post("http://httpbin.org/status/404")
    assert not res.ok
    assert res.status_code == 404

    with pytest.raises(HTTPError):
        res.raise_for_status()

    assert str(res) == "<Response [{code}]>".format(
        code = 404
    )


# imports - standard imports
import os

# imports - module imports
import pipupgrade
from   pipupgrade.util.types import auto_typecast

PREFIX = "%s" % pipupgrade.__name__.upper()

def getenvvar(name, prefix = PREFIX, seperator = "_"):
	if not prefix:
		prefix	  = ""
		seperator = ""

	envvar = "%s%s%s" % (prefix, seperator, name)
	return envvar

def getenv(name, default = None, cast = True, prefix = PREFIX, seperator = "_", raise_err = False):
    envvar = getenvvar(name, prefix = prefix, seperator = seperator)

    if not envvar in list(os.environ) and raise_err:
        raise KeyError("Environment Variable %s not found." % envvar)

    value  = os.getenv(envvar, default)
    value  = auto_typecast(value) if cast else value

    return value

def value_to_envval(value):
	"""
	Convert python types to environment values
	"""

	if not isinstance(value, str):
		if   value == True:
			value = "true"
		elif value == False:
			value = "false"
		elif isinstance(value, int):
			value = str(value)
		else:
			raise TypeError("Unknown parameter type %s with value %r" % (value, type(value)))

	return value


# imports - standard imports
from   	datetime	import datetime
from 	functools 	import partial
import 	os.path as osp
import 	re

# imports - module imports
from pipupgrade.__attr__    import __name__ as NAME
from pipupgrade 	 		import _pip, semver, request as req, db, log
from pipupgrade.tree 		import Node as TreeNode
from pipupgrade.util.string import kebab_case, strip
from pipupgrade._compat		import iteritems

logger  = log.get_logger()
_db		= db.get_connection() 

def _get_pypi_info(name, raise_err = True):
	url  = "https://pypi.org/pypi/{}/json".format(name)
	res  = req.get(url)

	info = None

	if res.ok:
		data = res.json()
		info = data["info"]
	else:
		if raise_err:
			res.raise_for_status()

	return info

def _get_pip_info(*args, **kwargs):
	packages	= args
	pip_exec	= kwargs.get("pip_exec", None)
	
	_, out, _	= _pip.call("show", *packages, pip_exec = pip_exec,
		output = True)
	results		= [strip(o) for o in out.split("---")]
	
	info		= dict()
	
	for i, package in enumerate(packages):
		result = results[i]

		detail = dict((kebab_case(k), v) \
			for k, v in \
				iteritems(
					dict([(s + [""]) if len(s) == 1 else s \
						for s in [re.split(r":\s?", o, maxsplit = 1) \
							for o in result.split("\n")]]
					)
				)
		)

		info[package] = detail
	
	return info

def _get_package_version(package, pip_exec = None):
	info = _get_pip_info(package, pip_exec = pip_exec)[package]
	return info["version"]

class Package:
	def __init__(self, package, sync = False, pip_exec = None):
		logger.info("Initializing Package %s of type %s..." % (package, type(package)))

		self.current_version = None

		if   isinstance(package, (_pip.Distribution, _pip.DistInfoDistribution,
			_pip.EggInfoDistribution)):
			self.name            = package.project_name
			self.current_version = package.version
		elif isinstance(package, _pip.InstallRequirement):
			self.name            = package.name

			if hasattr(package, "req"):
				if hasattr(package.req, "specifier"):
					self.current_version = str(package.req.specifier)
			else:
				self.current_version = package.installed_version
		elif isinstance(package, dict):
			self.name            = package["name"]
			self.current_version = package["version"]
			self.latest_version  = package.get("latest_version")
		elif isinstance(package, str):
			self.name			 = package
			if pip_exec:
				self.current_version = _get_package_version(package,
					pip_exec = pip_exec)

		res = None

		try:
			logger.info("Fetching package %s information from DB..." % self.name)

			res = _db.query("""
				SELECT *
				FROM `tabPackage`
				WHERE name = '%s'
			""" % self.name)
		except db.OperationalError as e:
			logger.warn("Unable to fetch package name. %s" % e)

		if not res or sync:
			logger.info("Fetching PyPI info for package %s..." % self)
			_pypi_info = _get_pypi_info(self.name, raise_err = False) or { }
		
			if not hasattr(self, "latest_version"):
				self.latest_version = _pypi_info.get("version")

			self.home_page = _pypi_info.get("home_page")

		if not res:
			try:
				logger.info("Attempting to INSERT package %s into database." % self)

				_db.query("""
					INSERT INTO `tabPackage`
						(name, latest_version, home_page, _created_at)
					VALUES
						('%s', '%s', '%s', '%s')
				""" % (self.name, self.latest_version, self.home_page, datetime.now()))
			except (db.IntegrityError, db.OperationalError) as e:
				logger.warn("Unable to save package name. %s" % e)
		else:
			if sync:
				logger.info("Attempting to UPDATE package %s within database." % self)

				try:
					_db.query("""
						UPDATE `tabPackage`
							SET latest_version = '%s', home_page = '%s', _updated_at = '%s'
						WHERE
							name = '%s'
					""" % (self.latest_version, self.home_page, datetime.now(), self.name))
				except db.OperationalError as e:
					logger.warn("Unable to update package name. %s" % e)
			else:
				logger.info("Using cached info for package %s." % self)

				self.latest_version = res["latest_version"]
				self.home_page      = res["home_page"]

		self.dependency_tree = TreeNode(self)

		try:
			self.difference = semver.difference(
				self.current_version,
				self.latest_version
			)
		except (TypeError, ValueError):
			self.difference = None

	def __repr__(self):
		repr_ = "<Package %s%s>" % (self.name,
			" (%s)" % self.current_version if self.current_version else "")
		return repr_

	def to_dict(self):
		dependencies = self.dependency_tree.to_dict(repr_ = lambda x: x.name)
		
		dict_ 		 = dict({
					   "name": self.name,
			"current_version": self.current_version,
			 "latest_version": self.latest_version,
				  "home_page": self.home_page,
			   "dependencies": dependencies[self.name]
		})

		return dict_


# pylint: disable=E1101

# imports - compatibility imports
from pipupgrade._compat import _is_python_version

# imports - standard imports
import itertools

def compact(arr, type_ = list):
    return type_(filter(bool, arr))

def squash(seq):
    value = seq

    if isinstance(value, (list, tuple)) and len(value) == 1:
        value = value[0]
    
    return value

def flatten(arr):
    if _is_python_version(major = 2, minor = 6):
        chainer = itertools.chain.from_iterable
    else:
        chainer = itertools.chain

    flattened = list(chainer(*arr))

    return flattened

def sequencify(value, type_ = list):
    if not isinstance(value, (list, tuple)):
        value = list([value])

    value = type_(value)
        
    return value


# imports - standard imports
from   contextlib      import contextmanager
import multiprocessing as mp
from   multiprocessing.pool import Pool

class NoDaemonProcess(mp.Process):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass

class NoDaemonProcessPool(Pool):
    Process = NoDaemonProcess

@contextmanager
def pool(class_ = Pool, *args, **kwargs):
    pool = class_(*args, **kwargs)
    yield pool
    pool.terminate()

@contextmanager
def no_daemon_pool(*args, **kwargs):
    with pool(class_ = NoDaemonProcessPool, *args, **kwargs) as p:
        yield p

        p.close(); p.join()


# imports - module imports
from pipupgrade.semver import parse, difference

# imports - test imports
import pytest

def test_parse():
    version = parse("1.2.3")
    assert version["major"] == 1
    assert version["minor"] == 2
    assert version["patch"] == 3
    assert not version["prerelease"]
    assert not version["build"]

    version = parse("1.2.3-develop+1234")
    assert version["major"] == 1
    assert version["minor"] == 2
    assert version["patch"] == 3
    assert version["prerelease"] == "develop"
    assert version["build"]      == "1234"

    with pytest.raises(ValueError):
        parse("foobar")
    
    with pytest.raises(ValueError):
        parse("1.0")

def test_difference():
    assert difference("1.2.3", "2.3.4") == "major"
    assert difference("1.2.3", "1.3.4") == "minor"
    assert difference("1.2.3", "1.2.1") == "patch"
    assert difference("1.2.3", "1.2.3") == None

    with pytest.raises(ValueError):
        difference("1.2.3", "1.0")

    with pytest.raises(ValueError):
        difference("1.2.3.4", "1.2.3.4")

    with pytest.raises(NotImplementedError):
        difference("1.2.3-develop", "1.2.3")


# imports - module imports
from pipupgrade.util.types import (
    get_function_arguments
)

def test_get_function_arguments():
    def foobar(foo = "bar", bar = "baz"):
        pass
    def barfoo():
        pass
    foobar(); barfoo() # Increase coverage
    
    assert get_function_arguments(foobar) == dict(foo = "bar", bar = "baz")
    assert get_function_arguments(barfoo) == dict()


# imports - standard imports
import collections

def merge_dict(*args):
    merged = dict()

    for arg in args:
        copy = arg.copy()
        merged.update(copy)

    return merged

def dict_from_list(keys, values):
    return dict(zip(keys, values))

def autodict():
    _autodict = collections.defaultdict(autodict)
    return _autodict


# imports - compatibility imports
from pipupgrade._compat import iteritems

# imports - standard imports
import pip
import json

# imports - module imports
from pipupgrade.util.system  import which, popen
from pipupgrade.util.string  import kebab_case
from pipupgrade.util.environ import value_to_envval
from pipupgrade.util.array   import sequencify
from pipupgrade              import log

logger = log.get_logger()

PIP9 = int(pip.__version__.split(".")[0]) < 10

if PIP9:
    # from pip                 import get_installed_distributions
    from pip.req             import parse_requirements
    from pip.req.req_install import InstallRequirement
else:
    # from pip._internal.utils.misc      import get_installed_distributions
    from pip._internal.req             import parse_requirements
    from pip._internal.req.req_install import InstallRequirement

from pip._vendor.pkg_resources import (
    Distribution,
    DistInfoDistribution,
    EggInfoDistribution
)

def _get_pip_executable(multiple = False):
    pips  = ("pip", "pip3", "pip2")
    execs = [ ]

    for pip_ in pips:
        exec_ = which(pip_)
        if exec_:
            if not multiple:
                return exec_
            else:
                if exec_ not in execs:
                    execs.append(exec_)

    if not execs:
        raise ValueError("pip executable not found.")

    return execs

_PIP_EXECUTABLE  = _get_pip_executable()
_PIP_EXECUTABLES = _get_pip_executable(multiple = True)

def call(*args, **kwargs):
    pip_exec  = kwargs.pop("pip_exec", _PIP_EXECUTABLE)
    quiet     = kwargs.pop("quiet",  False)
    output    = kwargs.pop("output", False)
    raise_err = kwargs.pop("raise_err", True)

    params    = sequencify(pip_exec) + sequencify(args)
    
    for flag, value in iteritems(kwargs):
        if value != False:
            flag  = "--%s" % kebab_case(flag, delimiter = "_")
            params.append(flag)

            if not isinstance(value, bool):
                value = value_to_envval(value)
                params.append(value)

    output = output or quiet
	
    output = popen(*params, output = output, raise_err = raise_err)
    
    return output


# imports - standard imports
from pipupgrade.cli.parser import get_args

def test_args():
    def _get_arg(args, arg):
        if hasattr(args, arg):
            return getattr(args, arg)
        else:
            return args[arg]
    
    def _assert_args(args):
        assert _get_arg(args, "yes")      == False
        assert _get_arg(args, "check")    == False
        assert _get_arg(args, "latest")   == False
        assert _get_arg(args, "no_color") == False

    args = get_args()
    _assert_args(args)

    args = get_args(as_dict = False)
    _assert_args(args)

    args = ["--verbose"]
    args = get_args(args, known = False)
    _assert_args(args)


# imports - standard imports
import subprocess

# imports - test imports
import pytest

# imports - module imports
from pipupgrade import _pip

def test_imports():
    from pipupgrade._pip import (
        parse_requirements          as _,
        InstallRequirement          as _,
        # get_installed_distributions as _,
        Distribution                as _,
        DistInfoDistribution        as _,
        EggInfoDistribution         as _
    )

def test_call(tmpdir):
    def assert_pip_call(response, code = 0, output = True, err = False):
        rcode, rout, rerr = response
        assert rcode == code

        def _assert_outerr(routerr, outerr):
            if isinstance(output, bool):
                if output:
                    assert rout
                else:
                    assert not rout
            else:
                assert rout == output

        _assert_outerr(rout, output)
        _assert_outerr(rerr, err)

    directory = tmpdir.mkdir("tmp")
    tempfile  = directory.join("tmp.log")
    path      = str(tempfile)

    _pip.call("install", "pipupgrade")
    assert_pip_call(_pip.call("install", "pipupgrade", quiet = True))
    
    _pip.call("install", "pipupgrade", log = path)
    assert tempfile.read()

    # assert_pip_call(_pip.call("list", output = True))


# imports - standard imports
import re

_REGEX_ANSI_ESCAPE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")

def strip(string):
    string = string.lstrip()
    string = string.rstrip()

    return string 

def strip_ansi(string):
    string = _REGEX_ANSI_ESCAPE.sub("", string)
    return string

def pluralize(string, count = 1):
    # A very shitty pluralizer
    if not string.endswith("s"):
        if count > 1:
            string += "s"
    
    return string

def kebab_case(string, delimiter = " "):
    words = string.replace(delimiter, " ").split()
    kebab = "-".join([word.lower() for word in words])
    
    return kebab

def safe_encode(obj, encoding = "utf-8"):
    try:
        obj = obj.encode(encoding)
    except (AttributeError, UnicodeEncodeError):
        pass
    
    return obj

def safe_decode(obj, encoding = "utf-8"):
    try:
        obj = obj.decode(encoding)
    except (AttributeError, UnicodeEncodeError):
        pass
    
    return obj


# imports - module imports
from pipupgrade.util.string import (strip, strip_ansi, pluralize, kebab_case,
    safe_encode, safe_decode)
from pipupgrade import cli

def test_strip():
    string = "foobar"
    assert strip(string) == string

    string = "\n   foobar\nfoobar   \n   "
    assert strip(string) == "foobar\nfoobar"

    string = "\n\n\n"
    assert strip(string) == ""

def test_strip_ansi():
    assert strip_ansi(cli.format("foobar", cli.GREEN)) == "foobar"
    assert strip_ansi(cli.format("barfoo", cli.BOLD))  == "barfoo"

def test_pluralize():
    assert pluralize("package",  1) == "package"
    assert pluralize("package",  2) == "packages"
    assert pluralize("packages", 2) == "packages"

def test_kebab_case():
    assert kebab_case("foo bar") == "foo-bar"
    assert kebab_case("Foo Bar") == "foo-bar"
    assert kebab_case("FOO BAR") == "foo-bar"

    assert kebab_case("_FOO_BAR_", delimiter = "_") == "foo-bar"
    assert kebab_case("foo_bar",   delimiter = "_") == "foo-bar"

def test_safe_encode():
    assert safe_encode(b"foobar") == b"foobar"
    assert safe_encode( "foobar") == b"foobar"

    assert safe_encode(123456789) == 123456789

def test_safe_decode():
    assert safe_decode(b"foobar") == "foobar"
    assert safe_decode( "foobar") == "foobar"
    
    assert safe_decode(123456789) == 123456789


# imports - standard imports
import os.path as osp
import subprocess

# imports - test imports
import pytest

# imports - module imports
from pipupgrade.__attr__ import (
    read,
    pardir,
    strip,
    safe_decode,
    sequence_filter,
    get_revision
)

def call(*args, **kwargs):
    subprocess.call(args, **kwargs)

def test_read(tmpdir):
    directory = tmpdir.mkdir("tmp")
    tempfile  = directory.join("foobar.txt")
    tempfile.write("foobar")

    assert tempfile.read() == read(str(tempfile))

    tempfile  = directory.join("barfoo.txt")
    tempfile.write(\
        """
        foobar
        \n
        barfoo
        """
    )

    assert tempfile.read() == read(str(tempfile))

def test_pardir():
    assert pardir(__file__)    == osp.dirname(__file__)
    assert pardir(__file__, 2) == osp.dirname(osp.dirname(__file__))

def test_strip():
    string = "foobar"
    assert strip(string) == string

    string = "\n   foobar\nfoobar   \n   "
    assert strip(string) == "foobar\nfoobar"

    string = "\n\n\n"
    assert strip(string) == ""

    string = "\r\nfoobar\nfoobar\n"
    assert strip(string) == "foobar\nfoobar"

def test_safe_decode():
    assert safe_decode(b"foobar") == "foobar"
    assert safe_decode( "foobar") == "foobar"
    
    assert safe_decode(123456789) == 123456789

def test_sequence_filter():
    assert sequence_filter([0,1,2,3,4,5], filter_ = lambda x: x % 2 == 0)                == [0,2,4]
    assert sequence_filter([0,1,2,3,4,5], filter_ = lambda x: x % 2 != 0, type_ = tuple) == (1,3,5)

def test_get_revision(tmpdir):
    directory = tmpdir.mkdir("tmp")
    path      = str(directory)
    
    with pytest.raises(subprocess.CalledProcessError):
        get_revision(path)
    
    assert get_revision(path, raise_err = False) == None

    # Initialize the git repository
    call("git","init",path)
    call("git","config","user.email","foobar@foobar.com", cwd = path)
    call("git","config","user.name" ,"Foo Bar", cwd = path)

    with pytest.raises(subprocess.CalledProcessError):
        get_revision(path)

    tempfile  = directory.join("foobar.txt")
    tempfile.write("foobar")

    call("git","add",".", cwd = path)
    call("git","commit","-m","'Test Commit'", cwd = path)

    assert len(get_revision(path))               == 40
    assert len(get_revision(path, short = True)) == 7


# imports - standard imports
import os.path      as osp

# imports - module imports
from pipupgrade             import __name__ as NAME, __version__
from pipupgrade.util.system import pardir, makedirs, touch
from pipupgrade.util.types  import auto_typecast
from pipupgrade.util._dict  import autodict
from pipupgrade._compat     import iteritems, configparser

PATH            = autodict()
PATH["BASE"]    = pardir(__file__)
PATH["DATA"]    = osp.join(PATH["BASE"], "data")
PATH["CACHE"]   = osp.join(osp.expanduser("~"), ".%s" % NAME)

class Configuration:
    def __init__(self, location = PATH["CACHE"], name = "config"):
        self.location   = location
        makedirs(self.location, exist_ok = True)
        
        self.name       = "%s.ini" % name
        
    @property
    def config(self):
        path    = osp.join(self.location, self.name)
        config  = cp.ConfigParser()
        
        if osp.exists(path):
            config.read(path)

        return config

    def get(self, section, key):
        config = self.config

        if not section in config:
            raise KeyError("No section %s found." % section)

        if not key in config[section]:
            raise KeyError("No key %s found." % key)
        
        value = auto_typecast(config[section][key])

        return value
        
    def set(self, section, key, value, force = False):
        config = self.config

        if not section in config:
            config[section] = dict({ key: value })
        else:
            if not key in config[section]:
                config[section][key] = value
            else:
                if force:
                    config[section][key] = value

        path = osp.join(self.location, self.name)
        with open(path, "w") as f:
            config.write(f)
        
class Settings:
    _DEFAULTS = {
        "version": __version__
    }

    def __init__(self):
        self.config = Configuration()

        self._init()

    def _init(self):
        for k, v in iteritems(Settings._DEFAULTS):
            self.set(k, v)

    def get(self, key):
        return self.config.get("settings", key)

    def set(self, key, value):
        self.config.set("settings", key, value)


# imports - standard imports
import subprocess as sp

# imports - module imports
from pipupgrade.util.system import popen
from pipupgrade.exception   import (
    PipupgradeError,
    PopenError
)

# imports - test imports
import pytest

def test_pipupgrade_error():
    with pytest.raises(PipupgradeError):
        raise PipupgradeError

def test_popen_error():
    with pytest.raises(PopenError):
        popen("python -c 'raise TypeError'")

    assert isinstance(
        PopenError(0, "echo foobar"),
        (PipupgradeError, sp.CalledProcessError)
    )
    assert isinstance(PipupgradeError(), Exception)


# imports - test imports
import pytest

# imports - module imports
from pipupgrade.tree import Node

def test_node():
    tree1 = Node("foo")
    assert tree1.empty == True
    assert tree1 == Node("foo")

    assert str(tree1) == "<Node 'foo'>"

    assert tree1.render() == \
"""\
foo
"""

    tree2 = Node("foo", children = ["bar", "baz"])
    assert tree2.parent == None
    assert tree2 != Node("foo", children = ["bar", "baz", "boo"])

    assert tree2.render() == \
"""\
foo
  bar
  baz
"""

    assert tree2.render(indent = 4) == \
"""\
foo
    bar
    baz
"""

    assert tree2.render(indent = 4, formatter = lambda x: "* %s" % x) == \
"""\
* foo
    * bar
    * baz
"""

    assert      tree2.find(lambda x: x.obj == "foo")
    assert      tree2.find(lambda x: x.obj == "bar")
    assert not  tree2.find(lambda x: x.obj == "foobaz")

    # tree3 = Node("foo")
    # tree3.add_children(["bar", "baz"])

    # tree4           = Node("foo")
    # tree4.children  = ["bar", "baz"]

    # with pytest.raises(TypeError):
    #     tree4.children = "foo"


# imports - compatibility imports
from __future__ import print_function
from pipupgrade._compat import input

# imports - standard imports
import sys, os
import platform
import inspect

# imports - module imports
from pipupgrade._compat      import EX_OK
from pipupgrade.util.imports import import_handler

_ACCEPTABLE_INPUTS_YES      = ("", "y", "Y")
_ACCEPTABLE_INPUTS_QUIT     = ("q", "Q")

# https://gist.github.com/RDCH106/6562cc7136b30a5c59628501d87906f7
_CAN_ANSI_FORMAT_WINDOWS    = (
    os.name == "nt" \
        and platform.release() == "10" \
        and platform.version() >= "10.0.14393"
)
_CAN_ANSI_FORMAT            = (
    # check if output is a terminal
    sys.stdout.isatty() \
        # check if stdin and stdout are the same
        or os.fstat(0) == os.fstat(1)
) or _CAN_ANSI_FORMAT_WINDOWS

_ANSI_FORMAT = "\033[{}m"
_format_ansi = lambda x: _ANSI_FORMAT.format(x)

BOLD      = _format_ansi("0;1")
RED       = _format_ansi("0;91")
GREEN     = _format_ansi("0;92")
YELLOW    = _format_ansi("0;93")
CYAN      = _format_ansi("0;96")
ORANGE    = _format_ansi("0;40")
CLEAR     = _format_ansi("0")

def confirm(query, quit_ = True):
    choices = "[Y/n%s]" % "/q" if quit_ else ""
    query   = "%s %s: " % (query, choices)

    output  = input(query)

    if output in _ACCEPTABLE_INPUTS_QUIT:
        sys.exit(EX_OK)
    
    return output in _ACCEPTABLE_INPUTS_YES

def format(string, type_):
    if _CAN_ANSI_FORMAT_WINDOWS:
        kernel32 = import_handler("ctypes.windll.kernel32")
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

    if _CAN_ANSI_FORMAT:
        string = "{}{}{}".format(type_, string, CLEAR)

    return string

def echo(string = "", nl = True):
    print(string, end = "\n" if nl else "")


# imports - module imports
from    pipupgrade import cli

# imports - test imports
import  pytest
from    testutils import assert_input, assert_stdout

def test_confirm(capfd):
    query  = "foobar"
    stdout = "{} [Y/n/q]: ".format(query)

    def _assert_confirm(stdout):
        assert_input(capfd, query, "Y", expected = True,  input_ = cli.confirm, stdout = stdout)
        assert_input(capfd, query, "y", expected = True,  input_ = cli.confirm, stdout = stdout)
        assert_input(capfd, query,"\n", expected = True,  input_ = cli.confirm, stdout = stdout)
        assert_input(capfd, query, "n", expected = False, input_ = cli.confirm, stdout = stdout)
        assert_input(capfd, query, "1", expected = False, input_ = cli.confirm, stdout = stdout)

    _assert_confirm(stdout)

    stdout = "{} [Y/n]: ".format(query)

    with pytest.raises(SystemExit):
        assert_input(capfd, query, "q", expected = False, input_ = cli.confirm, stdout = stdout)
    
    with pytest.raises(SystemExit):
        assert_input(capfd, query, "Q", expected = False, input_ = cli.confirm, stdout = stdout)

    # assert_input(capfd, query, "Y", expected = True,  input_ = cli.confirm, stdout = stdout, input_args = { 'quit_': False })
    # assert_input(capfd, query, "y", expected = True,  input_ = cli.confirm, stdout = stdout, input_args = { 'quit_': False })
    # assert_input(capfd, query,"\n", expected = True,  input_ = cli.confirm, stdout = stdout, input_args = { 'quit_': False })
    # assert_input(capfd, query, "n", expected = False, input_ = cli.confirm, stdout = stdout, input_args = { 'quit_': False })
    # assert_input(capfd, query, "1", expected = False, input_ = cli.confirm, stdout = stdout, input_args = { 'quit_': False })

def test_format():
    string = "foobar"

    def _assert_format(string, type_):
        assert cli.format(string, type_) == "{}{}{}".format(type_, string, cli.CLEAR)

    _assert_format(string, cli.GREEN)
    _assert_format(string, cli.RED)
    _assert_format(string, cli.BOLD)

def test_echo(capfd):
    query  = "foobar"
    cli.echo(query, nl = False)
    assert_stdout(capfd, query)
    
    cli.echo(query, nl = True)
    assert_stdout(capfd, "{}\n".format(query))


from pipupgrade.util.imports import HandlerRegistry, import_handler

def test_handler_registry():
    registry = HandlerRegistry()

    assert len(registry) == 0

    registry["os"]

    assert "os" in registry and __import__("os") == registry["os"]
    assert len(registry) == 1

def test_import_handler():
    assert import_handler("os")         == __import__("os")
    assert import_handler("pipupgrade") == __import__("pipupgrade")
    
    assert import_handler("pipupgrade.util.imports.import_handler") == import_handler


# imports - standard imports
import collections

# imports - module imports
from pipupgrade._compat import (cmp, input, StringIO, iteritems, iterkeys,
    itervalues, PYTHON_VERSION, _is_python_version)

# imports - test imports
from testutils import mock_input, assert_input

def test_imports():
    from pipupgrade._compat import (
        StringIO    as _,

        urlopen     as _,
        Request     as _,
        urlencode   as _,
        
        HTTPError   as _,

        zip         as _,
        zip_longest as _,
        
        input       as _,

        EX_OK       as _,
        EX_NOINPUT  as _,

        ModuleNotFoundError as _
    )

def test__is_python_version():
    def _assert_version(major, minor):
        if PYTHON_VERSION.major == major and PYTHON_VERSION.minor == minor:
            assert _is_python_version(major = major, minor = minor)

    _assert_version(2, 7)
    _assert_version(3, 4)
    _assert_version(3, 5)
    _assert_version(3, 6)
    _assert_version(3, 7)

    assert _is_python_version(
        major = PYTHON_VERSION.major,
        minor = PYTHON_VERSION.minor,
        patch = PYTHON_VERSION.micro
    )

def test_input(capfd):
    query = "foobar"
    assert_input(capfd, query, "Y", input_ = input)

def test_cmp():
    assert cmp(1, 2) == -1
    assert cmp(1, 1) ==  0
    assert cmp(2, 1) ==  1

def test_iteritems():
    dict_ = dict(foo = "bar")
    
    assert isinstance(iteritems(dict_), collections.Iterable)

    for k, v in iteritems(dict_):
        assert dict_[k] == v

def test_iterkeys():
    dict_ = dict(foo = "bar")
    
    assert isinstance(iterkeys(dict_), collections.Iterable)

    for k in iterkeys(dict_):
        assert k in dict_

def test_itervalues():
    dict_ = dict(foo = "bar")
    
    assert isinstance(itervalues(dict_), collections.Iterable)

    assert list(itervalues(dict_)) == ["bar"]


def _render_tree(node, depth = 0, indent = 2, formatter = None):
    spacing     = (indent * " ") * depth
    formatted   = formatter(node.obj) if formatter else node.obj

    string      = "%s%s\n" % (spacing, formatted)

    for child in node.children:
        string += _render_tree(child, depth + 1,
            indent = indent, formatter = formatter)
    
    return string

def _check_node(node, query):
    result = False

    if callable(query):
        result = bool(query(node))
    else:
        result = node == query

    return result

class Node:
    def __init__(self, obj, children = [ ], parent = None):
        self.obj        = obj
        self._children  = [ ]

        for child in children:
            self.add_child(child)

        self._parent    = parent

    @property
    def parent(self):
        return getattr(self, "_parent", None)

    @parent.setter
    def parent(self, value):
        if self.parent == value:
            pass
        else:
            self._parent = value

    @property
    def children(self):
        return getattr(self, "_children", [ ])

    @children.setter
    def children(self, value):
        if self.children == value:
            pass
        elif not isinstance(value, (list, tuple)):
            raise TypeError("Children must be of type (list, tuple), found %s"
                % type(value)
            )
        else:
            children = [ ]

            for child in value:
                if not isinstance(child, Node):
                    child = Node(child, parent = self)

                children.append(child)

            self._children = children

    def __eq__(self, other):
        equals = False

        if isinstance(other, Node):
            equals = self.find(lambda x: x.obj != other.obj)
        
        return equals
    
    @property
    def empty(self):
        nchildren = len(self.children)
        return nchildren == 0

    def add_child(self, child):
        if not isinstance(child, Node):
            child = Node(child, parent = self)

        self.children.append(child)

    def add_children(self, *children):
        for child in children:
            self.add_child(child)

    def __repr__(self):
        repr_ = "<Node '%s'%s>" % (str(self.obj), 
            "parent='%s'" % self.parent if self.parent else "")
        return repr_

    def render(self, indent = 2, formatter = None):
        string = _render_tree(self, indent = indent, formatter = formatter)
        return string

    def to_dict(self, repr_ = None):
        key   = repr_(self.obj) if repr_ else str(self.obj)
        dict_ = dict({
            key: [d.to_dict(repr_ = repr_) \
                for d in self.children]
        })

        return dict_

    def find(self, query):
        """
        Performs a Depth-First Search to find a Node based on a query provieded.
        """
        found = None

        if _check_node(self, query):
            found = self

        if not found:
            for child in self.children:
                found = child.find(query)
                if found:
                    break

        return found


# imports - compatibility imports
from pipupgrade._compat import (
    urlopen,
    Request,
    urlencode,
    HTTPError,
    ModuleNotFoundError
)

# imports - standard imports
import json

# imports - module imports
from pipupgrade.request.response import Response
from pipupgrade.util.string      import safe_encode
from pipupgrade.log              import get_logger

logger = get_logger()

# YAGNI: This patched "requests" works only for pipupgrade's use-cases.

def get(*args, **kwargs):
    logger.info("Dispatching GET request with arguments %s and parameters %s." % (
        args, kwargs
    ))

    try:
        import requests as req
        return req.get(*args, **kwargs)
    except (ImportError, ModuleNotFoundError):
        url      = args[0]
        response = Response()

        try:
            http_response    = urlopen(url)
            status_code      = http_response.getcode()

            response.content = http_response.read()

            http_response.close()
        except HTTPError as e:
            status_code      = e.getcode()

        response.url         = url
        response.status_code = status_code

        return response

def post(*args, **kwargs):
    logger.info("Dispatching POST request with arguments %s and parameters %s." % (
        args, kwargs
    ))

    try:
        import requests as req
        return req.post(*args, **kwargs)
    except (ImportError, ModuleNotFoundError):
        url      = args[0]
        data     = kwargs.get("data",    { })
        headers  = kwargs.get("headers", { })

        response = Response()
        
        try:
            data             = safe_encode(urlencode(data))
            request          = Request(url, data = data, headers = headers)
            http_response    = urlopen(request)
            status_code      = http_response.getcode()

            response.content = http_response.read()

            http_response.close()
        except HTTPError as e:
            status_code      = e.getcode()

        response.url         = url
        response.status_code = status_code

        return response